import os
import tempfile
from abc import ABC, abstractmethod
from collections import OrderedDict
from contextlib import contextmanager
from typing import Any, Dict, Generic, Iterator, List, Optional, Tuple, TypeVar, Union

from encode_utils.connection import Connection

from accession.database.connection import DbSession
from accession.database.models import Run

K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T")
U = TypeVar("U", bound="PreferredDefaultFilePatch")


class LruCache(Generic[K, V]):
    """
    Helper class implementing a LRU cache using an `OrderedDict`. The Generic class it
    inherits from is purely for type checking. `K` is the type variable for the key,
    and `V` is the type variable for the value. If you try to insert mixed types, mypy
    will complain.
    """

    def __init__(self, max_size: int = 128) -> None:
        """
        `max_size` is an upper bound on the maximum size of the cache. When the cache is
        at this size, insertions will result in eviction of the oldest values in the
        cache.
        """
        self.max_size = max_size
        self.data: OrderedDict[K, V] = OrderedDict()

    def get(self, key: K) -> Optional[V]:
        """
        Get the value in the cache corresponding to the the key. Has the side effect of
        moving the key to the end of the cache to indicate it was recently used.
        """
        value = self.data.get(key)
        if value is not None:
            self.data.move_to_end(key)
        return value

    def insert(self, key: K, value: V) -> None:
        """
        Add an item to the cache. If the cache is at capacity then the oldest value will
        be cleared from the cache. If the key is already in the cache, the the value
        will be updated with the new value, and the item will moved to the end.
        """
        if len(self.data) == self.max_size:
            self.data.popitem(last=False)
        self.data[key] = value

    def invalidate(self, key: K) -> None:
        """
        Delete a given key from the cache. Not currently used in the production code
        but could be useful in the future, so will keep for now.
        """
        if key in self.data.keys():
            del self.data[key]


class PreferredDefaultFilePatch:
    PROFILE_KEY = "file"

    def __init__(self, at_id: str, qc_value: Union[int, float]) -> None:
        self.at_id = at_id
        self.qc_value = qc_value

    def __eq__(self, other: "U") -> bool:  # type: ignore[override]
        return all((self.at_id == other.at_id, self.qc_value == other.qc_value))

    def get_portal_patch(self) -> Dict[str, Union[str, bool]]:
        return {
            "preferred_default": True,
            Connection.PROFILE_KEY: self.PROFILE_KEY,
            Connection.ENCID_KEY: self.at_id,
        }


class AbstractRecorder(ABC):
    @abstractmethod
    def record(self, run: Run) -> None:
        raise NotImplementedError


class Recorder(AbstractRecorder):
    def __init__(self, use_in_memory_db: bool = False) -> None:
        if use_in_memory_db:
            self.db_session = DbSession.with_in_memory_db()
        else:
            self.db_session = DbSession.with_home_dir_db_path()

    def record(self, run: Run) -> None:
        self.db_session.session.add(run)
        self.db_session.session.commit()


def string_to_number(string: str) -> Union[float, str, int]:
    if not isinstance(string, str):
        return string
    try:
        return int(string)
    except Exception:
        try:
            return float(string)
        except Exception:
            return string


def flatten(nested_input: List[Any]) -> List[Any]:
    """Flattens a nested list.
    Args:
        input_list: A (possibly) nested list.
    Returns:
        A flattened list, preserving order.
    """

    if not nested_input:
        return []
    if isinstance(nested_input[0], list):
        return flatten(nested_input[0]) + flatten(nested_input[1:])
    else:
        return nested_input[:1] + flatten(nested_input[1:])


@contextmanager
def impersonate_file(data: bytes) -> Iterator[str]:
    """With this contextmanager one can use bytes or string as if it is a file.
    Usage:
        with impersonate_file(bytes_data) as filepath:
            function_expecting_file(filepath)
    """
    temporary_file = tempfile.NamedTemporaryFile(delete=False)
    temporary_file.write(data)
    temporary_file.close()
    try:
        yield temporary_file.name
    finally:
        os.unlink(temporary_file.name)


def get_api_keys_from_env() -> Tuple[str, str]:
    """
    Extracted from encode_utils.connection.Connection
    """
    api_key = os.environ["DCC_API_KEY"]
    secret_key = os.environ["DCC_SECRET_KEY"]
    return api_key, secret_key


def get_bucket_and_key_from_s3_uri(s3_uri: str) -> Tuple[str, str]:
    path_parts = s3_uri.replace("s3://", "").split("/")
    bucket = path_parts.pop(0)
    key = "/".join(path_parts)
    return bucket, key



import logging
from typing import List, Optional

import attr

from accession.encode_models import EncodeFile


@attr.s(auto_attribs=True)
class MatchingMd5Record:
    file_path: str
    portal_files: List[EncodeFile]


class PreflightHelper:
    """
    Encapsulates some helper functions used for doing dry run reporting, consumes data
    fed to it by the Accession instance and logs formatted messages.
    """

    def __init__(self, logger: logging.Logger):
        self.logger = logger

    def make_file_matching_md5_record(
        self, filename: str, matching: List[EncodeFile]
    ) -> Optional[MatchingMd5Record]:
        """
        Returns a record of all portal files with a matching md5sum, or None if no
        matching files were found.
        """
        if not matching:
            return None
        matching_md5_record = MatchingMd5Record(
            file_path=filename, portal_files=matching
        )
        return matching_md5_record

    def make_log_messages_from_records(
        self, matches: List[MatchingMd5Record]
    ) -> List[str]:
        header = (
            "File Path",
            "Matching Portal Files",
            "Portal Status",
            "Portal File Dataset",
        )
        rows = [header]
        for match in matches:
            for i, portal_file in enumerate(match.portal_files):
                if i == 0:
                    display_filename = match.file_path
                else:
                    display_filename = ""
                rows.append(
                    (
                        display_filename,
                        portal_file.accession,
                        portal_file.status.value,
                        portal_file.dataset,
                    )
                )
        columns = list(zip(*rows))
        column_widths = []
        for column in columns:
            lens = [len(i) for i in column]
            column_widths.append(max(lens))
        template = "{{:{}}} | {{:{}}} | {{:{}}} | {{:{}}}".format(*column_widths)
        messages = [template.format(*row) for row in rows]
        return messages

    def report_dry_run(self, matches: List[MatchingMd5Record]) -> None:
        """
        Print the report for the dry run. We use a format string to control the width
        for each column by adding padding where appropriate. The column width is
        governed by the length (in characters) of the longest element in each column.
        When there is more than one file at the portal that had a matching md5sum, the
        file path is not printed for subsequent report rows after the first match for
        visual clarity.
        """
        if not matches:
            self.logger.info("No MD5 conflicts found.")
            return
        else:
            self.logger.info("Found files with duplicate md5sums")
        messages = self.make_log_messages_from_records(matches)
        for msg in messages:
            self.logger.info(msg)



import datetime
import json
import os
import shutil
from base64 import b64decode
from pathlib import Path
from time import sleep
from typing import Callable, Iterator, Tuple
from unittest.mock import PropertyMock, create_autospec
from urllib.parse import urljoin

import encode_utils as eu
import pytest
import requests
from encode_utils.connection import Connection
from pytest_mock.plugin import MockerFixture
from WDL import parse_document

import docker
from accession import backends
from accession.accession import (
    Accession,
    AccessionChip,
    AccessionMicroRna,
    accession_factory,
)
from accession.analysis import Analysis
from accession.database.connection import DbSession
from accession.database.models import Run, RunStatus, WorkflowLabel
from accession.database.query import DbQuery
from accession.encode_models import (
    EncodeAttachment,
    EncodeCommonMetadata,
    EncodeDataset,
    EncodeDocument,
    EncodeDocumentType,
    EncodeFile,
)
from accession.file import GSFile
from accession.helpers import Recorder
from accession.metadata import FileMetadata
from accession.task import Task


@pytest.fixture(scope="session")
def local_encoded_server(
    api_credentials: Tuple[str, str], encoded_docker_image: str
) -> Iterator[str]:
    """
    Spin up the encoded local app (using default command and user) bound to host port
    8000, and kill the container when done.  Running the tests requires docker daemon to
    be running first. Note the session scope of the fixture, the app is expensive to
    spin up so we try to reuse it as many times as possible. This fixture is written as
    as generator so we can reenter the function stack frame to perform the clean up of
    killing the container.

    There are two conditions that could cause a retry here, handled in the same while
    loop. We expect that initially we'll get 504s, if nginx hasn't started yet, or 502s
    once nginx does start. Then, once we can hit the _indexer_state endpoint, we need to
    check that it's done indexing and check again later if it is not.

    Maybe we don't actually need to wait until it is done indexing, but I think it
    probably helps make the actual tests more stable.
    """
    host_port = 8000
    if not encoded_docker_image:
        raise ValueError(
            (
                "No docker image specfied, make sure you are using --docker [IMAGE] pytest "
                "argument"
            )
        )
    client = docker.from_env()
    try:
        container = client.containers.run(
            encoded_docker_image,
            detach=True,
            ports={"8000/tcp": host_port},
            auto_remove=True,
        )
    except Exception as e:
        raise RuntimeError(
            (
                f"Could not run docker image {encoded_docker_image}, make sure the Docker "
                "daemon is running and the image has been pulled"
            )
        ) from e
    server_address = f"localhost:{host_port}"
    retries_left = 6
    while True:
        sleep(10)
        if container.status == "exited":
            raise RuntimeError("Container unexpectedly exited early")
        try:
            response = requests.get(
                urljoin(f"http://{server_address}", "_indexer_state"),
                headers={"Accept": "application/json"},
                auth=api_credentials,
            )
        except requests.ConnectionError:
            retries_left -= 1
            continue
        if not response.ok:
            if retries_left > 0:
                retries_left -= 1
                continue
            container.kill()
            response.raise_for_status()
        res = response.json()
        if res.get("state", {}).get("status") == "done":
            break
        if retries_left > 0:
            retries_left -= 1
            continue
        container.kill()
        raise RuntimeError("Elasticsearch took too long to index")
    yield server_address
    container.kill()


@pytest.fixture(scope="session")
def db_session():
    return DbSession("sqlite:///:memory:", echo=True)


@pytest.fixture(scope="session")
def db_query(db_session):
    return DbQuery(db_session)


@pytest.fixture(scope="session")
def date():
    return datetime.date(1999, 9, 9)


@pytest.fixture(scope="session")
def run(db_session, date):
    label = WorkflowLabel(key="foo", value="bar")
    run = Run(
        experiment_at_id="ABC",
        workflow_id="123",
        date_created=date,
        status=RunStatus.Succeeded,
        workflow_labels=[label],
    )
    db_session.session.add(run)
    yield run
    db_session.session.delete(run)


@pytest.fixture(scope="session")
def additional_runs(db_session):
    run = Run(
        experiment_at_id="/experiments/ENCSR123ABC/",
        workflow_id="456",
        date_created=datetime.date(2020, 3, 10),
        status=RunStatus.Succeeded,
        workflow_labels=[WorkflowLabel(key="caper-str-label", value="bar")],
    )
    run2 = Run(
        experiment_at_id="/experiments/ENCSR456DEF/",
        workflow_id="60fb02a7-31ae-4e9e-8e77-9c2722189cb4",
        date_created=datetime.date(2020, 3, 11),
        status=RunStatus.Succeeded,
        workflow_labels=[WorkflowLabel(key="caper-str-label", value="baz")],
    )
    db_session.session.add(run)
    db_session.session.add(run2)
    yield
    db_session.session.delete(run)
    db_session.session.delete(run2)


@pytest.fixture
def server_name() -> str:
    return "mock_server.biz"


@pytest.fixture(scope="session")
def lab() -> str:
    return "/labs/encode-processing-pipeline/"


@pytest.fixture(scope="session")
def award() -> str:
    """
    This is the ENC4 award for the DCC
    """
    return "U24HG009397"


@pytest.fixture(scope="session")
def common_metadata(lab, award):
    return EncodeCommonMetadata(lab, award)


@pytest.fixture(scope="session")
def api_credentials() -> Tuple[str, str]:
    return ("foobar", "bazqux")


@pytest.fixture(scope="session")
def environ_api_credentials(api_credentials: Tuple[str, str]) -> None:
    key, secret = api_credentials
    os.environ["DCC_API_KEY"] = key
    os.environ["DCC_SECRET_KEY"] = secret


@pytest.fixture
def metadata_json_path() -> Path:
    current_dir = Path(__file__).resolve()
    metadata_json_path = current_dir.parent / "data" / "ENCSR609OHJ_metadata_2reps.json"
    return metadata_json_path


@pytest.fixture
def normal_analysis(
    mock_gc_backend, metadata_json_path: Path
) -> Analysis:  # noqa: F811
    normal_analysis = Analysis(
        FileMetadata(metadata_json_path), backend=mock_gc_backend
    )
    return normal_analysis


@pytest.fixture
def metadata_json(metadata_json_path: Path) -> Iterator[dict]:
    with open(metadata_json_path) as json_file:
        yield json.load(json_file)


@pytest.fixture
def mirna_replicated_metadata_path() -> str:
    current_dir = Path(__file__).resolve()
    metadata_json_path = current_dir.parent / "data" / "mirna_replicated_metadata.json"
    return str(metadata_json_path)


@pytest.fixture
def wdl_workflow():
    workflow = (
        'workflow myWorkflow { meta { version: "v1.2.3" } call myTask } task myTask {'
        " command { bar }}"
    )
    return workflow


@pytest.fixture
def encode_file_no_qc():
    return EncodeFile({"@id": "/files/foo/", "quality_metrics": []})


@pytest.fixture(scope="module")
def encode_attachment():
    return EncodeAttachment(
        filename="/my/dir/haz/my_text_file", contents=b"foo bar baz"
    )


@pytest.fixture(scope="module")
def encode_common_metadata():
    return EncodeCommonMetadata("/labs/lab/", "award")


@pytest.fixture
def encode_document(encode_attachment, encode_common_metadata):
    return EncodeDocument(
        encode_attachment,
        encode_common_metadata,
        EncodeDocumentType.WorkflowMetadata,
        aliases=["encode:foo"],
    )


@pytest.fixture
def gsfile(mocker):
    task = {
        "inputs": {
            "prefix": "pooled-pr1_vs_pooled-pr2",
            "fastqs_R1": ["gs://abc/spam.fastq.gz"],
        },
        "outputs": {},
    }
    my_task = Task("my_task", task)
    mocker.patch(
        "accession.file.GSFile.md5sum", mocker.PropertyMock(return_value="123")
    )
    return GSFile(key="foo", name="gs://abc/spam.fastq.gz", task=my_task, client="bar")


@pytest.fixture
def mock_file(mocker):
    file = mocker.Mock(
        filekeys=["foo"], filename="gs://foo/bar", size=123, md5sum="abc"
    )
    return file


@pytest.fixture
def mock_gs_file(mocker):
    # md5sum is yN0BGTic4eg+yn7K3BVlGw== in base 64
    file = mocker.MagicMock(
        filekeys=["foo"],
        filename="gs://foo/bar",
        size=3,
        md5sum="c8dd0119389ce1e83eca7ecadc15651b",
    )
    file.blob.download_as_string.return_value = b'{"foobar": "bazqux"}'
    return file


class LocalMockGsFile:
    SCHEME = "gs://"

    def __init__(self, key, name, task=None, used_by_task=None, client=None):
        self.filekeys = [key]
        self.filename = name
        self.task = task
        self.size = 3
        self.used_by_tasks = [used_by_task] if used_by_task else []
        self._md5_lookup = None

    @property
    def md5_lookup(self):
        if self._md5_lookup is None:
            md5_file = Path(__file__).resolve().parent / "data" / "gcloud_md5s.json"
            with open(md5_file) as f:
                self._md5_lookup = json.load(f)
        return self._md5_lookup

    @property
    def md5sum(self) -> str:
        b64_md5 = self.md5_lookup[self.filename]
        return b64decode(b64_md5).hex()

    def download_as_string(self) -> bytes:
        """
        We use the concatenation of the last two path elements (glob and filename) to
        find the file in the test data, which should be unique
        """
        file_name = self.filename.split("/")[-1]
        current_dir = Path(__file__).resolve()
        file_path = current_dir.parent / "data" / "files" / file_name
        with open(file_path, "rb") as f:
            binary_data = f.read()
        return binary_data

    def read_json(self):
        """
        Read file and convert to JSON
        """
        return json.loads(self.download_as_string().decode())

    def get_task(self) -> Task:
        """
        Either gets the `File`'s task or raises a ValueError.
        """
        task = self.task
        if task is None:
            raise ValueError("No task found")
        return task

    def get_uri_without_scheme(self) -> str:
        return self.filename.split(self.SCHEME)[-1]

    def get_filename_for_encode_alias(self) -> str:
        return self.get_uri_without_scheme().replace("/", "-")


@pytest.fixture
def mock_gc_backend(mocker):
    """
    A mocked instance of GCBackend that cannot initialize a Google Cloud client or make
    Google Cloud API calls.

    mocker fixture can only be used with function-scoped pytest fixtures for now:
    https://github.com/pytest-dev/pytest-mock/issues/136
    """
    mocker.patch("accession.file.GSFile", mock_gs_file)
    mocker.patch("accession.backends.GCBackend.client", mocker.PropertyMock())
    mock_gc_backend = backends.GCBackend()
    return mock_gc_backend


@pytest.fixture
def mock_accession_gc_backend(mocker):
    """
    Similar to mock_gc_backend, except it provides real md5sums and can access real test
    data files.
    """
    mocker.patch("accession.backends.GCBackend.client", mocker.PropertyMock())
    mocker.patch("accession.file.GSFile", LocalMockGsFile)
    mocker.patch("accession.backends.GSFile", LocalMockGsFile)
    mock_gc_backend = backends.GCBackend()
    return mock_gc_backend


@pytest.fixture
def accessioner_factory(
    mocker: MockerFixture,
    mock_accession_gc_backend,
    local_encoded_server: str,
    environ_api_credentials: None,
    lab: str,
    award: str,
) -> Callable:
    def _accessioner_factory(metadata_file: str, assay_name: str) -> Iterator:
        # Setup
        current_dir = Path(__file__).resolve()

        def mock_set_dcc_mode(self, dcc_mode: str) -> str:
            eu.DCC_MODES[dcc_mode] = {"host": dcc_mode, "url": f"http://{dcc_mode}/"}
            return dcc_mode

        mocker.patch.object(Connection, "_set_dcc_mode", mock_set_dcc_mode)

        accessioner = accession_factory(
            assay_name,
            metadata_file,
            local_encoded_server,
            lab,
            award,
            backend=mock_accession_gc_backend,
            no_log_file=True,
            use_in_memory_db=True,
        )

        mocker.patch.object(
            accessioner, "upload_file", return_value=None, autospec=True
        )
        with open(
            current_dir.parent / "data" / "validation" / f"{assay_name}" / "files.json"
        ) as f:
            expected_files = json.load(f)
        yield (accessioner, expected_files)
        shutil.rmtree(current_dir.parents[1] / "EU_Logs")

    return _accessioner_factory


@pytest.fixture
def mock_metadata(wdl_workflow):
    class MockMetadata:
        content = {"id": "foo", "workflowRoot": "gs://foo/bar", "calls": {}}
        workflow_id = "123"

        def get_parsed_workflow(self):
            return parse_document(wdl_workflow)

    return MockMetadata()


class MockAccessionSteps:
    path_to_json = "/dev/null"
    content = {"accession.steps": [{"a": "b"}]}
    quality_standard = "/quality-standards/bar/"


@pytest.fixture
def mock_accession_steps():
    return MockAccessionSteps()


@pytest.fixture
def mock_accession(
    mocker: MockerFixture,
    mock_accession_gc_backend,
    mock_metadata,
    mock_accession_steps: MockAccessionSteps,
    server_name: str,
    common_metadata: EncodeCommonMetadata,
) -> Accession:
    """
    Mocked accession instance with dummy __init__ that doesn't do anything and pre-baked
    assembly property. @properties must be patched before instantiation
    """
    mocker.patch.object(
        AccessionMicroRna, "assembly", new_callable=PropertyMock(return_value="hg19")
    )
    mocker.patch.object(
        AccessionMicroRna,
        "genome_annotation",
        new_callable=PropertyMock(return_value="V19"),
    )
    mocker.patch.object(
        Accession,
        "dataset",
        new_callable=PropertyMock(
            return_value=EncodeDataset(
                {
                    "@id": "/experiments/foo/",
                    "@type": ["Experiment", "Dataset", "Item"],
                    "assay_term_name": "microRNA",
                    "replicates": [
                        {"biological_replicate_number": 1},
                        {"biological_replicate_number": 2},
                    ],
                }
            )
        ),
    )
    mocker.patch.object(Accession, "preflight_helper", new_callable=PropertyMock())
    mocked_accession = AccessionMicroRna(
        mock_accession_steps,
        Analysis(mock_metadata, backend=mock_accession_gc_backend),
        create_autospec(Connection, dcc_url=server_name),
        common_metadata,
        Recorder(use_in_memory_db=True),
        no_log_file=True,
    )
    return mocked_accession


@pytest.fixture
def mock_accession_not_patched(
    mocker,
    mock_accession_gc_backend,
    mock_metadata,
    mock_accession_steps,
    server_name,
    common_metadata,
    mock_file,
):
    mocker.patch.object(
        Analysis, "raw_fastqs", new_callable=PropertyMock(return_value=[mock_file])
    )
    mock_accession = AccessionMicroRna(
        mock_accession_steps,
        Analysis(mock_metadata, backend=mock_accession_gc_backend),
        create_autospec(Connection, dcc_url=server_name),
        common_metadata,
        Recorder(use_in_memory_db=True),
        no_log_file=True,
    )
    return mock_accession


@pytest.fixture
def mock_accession_chip(
    mocker: MockerFixture,
    mock_accession_gc_backend,
    mock_metadata,
    mock_accession_steps: MockAccessionSteps,
    server_name: str,
    common_metadata: EncodeCommonMetadata,
) -> Accession:
    """
    Mocked accession instance with dummy __init__ that doesn't do anything and pre-baked
    assembly property. @properties must be patched before instantiation
    """
    mocker.patch.object(
        AccessionChip, "assembly", new_callable=PropertyMock(return_value="hg19")
    )
    mocker.patch.object(
        AccessionChip,
        "genome_annotation",
        new_callable=PropertyMock(return_value="V19"),
    )
    mocker.patch.object(
        Accession,
        "dataset",
        new_callable=PropertyMock(
            return_value=EncodeDataset(
                {
                    "@id": "foo",
                    "assay_term_name": "TF ChIP-seq",
                    "replicates": [
                        {"biological_replicate_number": 1},
                        {"biological_replicate_number": 2},
                    ],
                }
            )
        ),
    )
    mocked_accession = AccessionChip(
        mock_accession_steps,
        Analysis(mock_metadata, backend=mock_accession_gc_backend),
        server_name,
        common_metadata,
        Recorder(use_in_memory_db=True),
        no_log_file=True,
    )
    return mocked_accession


@pytest.fixture
def mock_accession_unreplicated(
    mocker: MockerFixture,
    mock_accession_gc_backend,
    mock_metadata,
    lab: str,
    award: str,
) -> Accession:
    """
    Mocked accession instance with dummy __init__ that doesn't do anything and pre-baked
    assembly property. @properties must be patched before instantiation
    """
    mocker.patch.object(
        Accession,
        "dataset",
        new_callable=PropertyMock(
            return_value=EncodeDataset(
                {
                    "@id": "foo",
                    "assay_term_name": "microRNA",
                    "replicates": [
                        {"biological_replicate_number": 1, "status": "released"}
                    ],
                }
            )
        ),
    )
    mocked_accession = AccessionMicroRna(
        "imaginary_steps.json",
        Analysis(mock_metadata, backend=mock_accession_gc_backend),
        "mock_server.biz",
        EncodeCommonMetadata(lab, award),
        Recorder(use_in_memory_db=True),
        no_log_file=True,
    )
    return mocked_accession



import pytest

from accession.file import GSFile, LocalFile, S3File


@pytest.fixture
def local_file():
    return LocalFile(key="foo", name="bar")


@pytest.fixture
def test_file_contents():
    return b"foo bar baz"


@pytest.fixture
def test_file(tmp_path, test_file_contents):
    path = tmp_path / "test_file.txt"
    path.write_bytes(test_file_contents)
    return LocalFile(key="foo", name=str(path))


@pytest.fixture
def gs_file_with_name_and_bucket(mocker):
    blob = mocker.Mock()
    blob.name = "baz"
    blob.bucket.name = "bar"
    mocker.patch("accession.file.GSFile.blob", mocker.PropertyMock(return_value=blob))
    file = GSFile(key="foo", name="gs://bar/baz")
    return file


@pytest.fixture
def auth(mocker):
    mocker.patch.dict("os.environ", {"DCC_API_KEY": "foo", "DCC_SECRET_KEY": "bar"})


def stub_download_as_string(start=None, end=None, data=b"abc123"):
    if start is None:
        start = 0
    if end is None or end + 1 >= len(data):
        end = len(data) - 1
    if start > end:
        return data
    return data[start : end + 1]


@pytest.fixture
def readable_gs_file(mocker):
    blob = mocker.Mock(download_as_string=stub_download_as_string, size=6)
    mocker.patch("accession.file.GSFile.blob", mocker.PropertyMock(return_value=blob))
    gs_file = GSFile(key="foo", name="gs://bar/baz")
    return gs_file


def test_gs_file_md5sum(mocker):
    blob = mocker.Mock(md5_hash="rQproAprtP/Prr1eFZpaHA==")
    mocker.patch("accession.file.GSFile.blob", mocker.PropertyMock(return_value=blob))
    gs_file = GSFile(key="foo", name="gs://bar/baz")
    assert gs_file.md5sum == "ad0a6ba00a6bb4ffcfaebd5e159a5a1c"


def test_gs_file_size(mocker):
    blob = mocker.Mock(size=3)
    mocker.patch("accession.file.GSFile.blob", mocker.PropertyMock(return_value=blob))
    gs_file = GSFile(key="foo", name="gs://bar/baz")
    assert gs_file.size == 3


def test_gs_file_b64_to_hex():
    assert (
        GSFile.b64_to_hex("rQproAprtP/Prr1eFZpaHA==")
        == "ad0a6ba00a6bb4ffcfaebd5e159a5a1c"
    )


def test_gs_file_get_uri_without_scheme(gs_file_with_name_and_bucket):
    assert gs_file_with_name_and_bucket.get_uri_without_scheme() == "bar/baz"


def test_gs_file_get_filename_for_encode_alias(gs_file_with_name_and_bucket):
    assert gs_file_with_name_and_bucket.get_filename_for_encode_alias() == "bar-baz"


def test_gs_file_read_bytes(mocker):
    blob = mocker.Mock()
    blob.download_as_string.return_value = b"foo"
    mocker.patch("accession.file.GSFile.blob", mocker.PropertyMock(return_value=blob))
    gs_file = GSFile(key="foo", name="gs://bar/baz")
    result = gs_file.read_bytes()
    assert result == b"foo"


def test_gs_file_read_json(mocker):
    blob = mocker.Mock()
    blob.download_as_string.return_value = b'{"foo":"bar"}'
    mocker.patch("accession.file.GSFile.blob", mocker.PropertyMock(return_value=blob))
    gs_file = GSFile(key="foo", name="gs://bar/baz")
    result = gs_file.read_json()
    assert result == {"foo": "bar"}


@pytest.mark.parametrize(
    "start,end,expected",
    [(None, None, b"abc123"), (None, 3, b"abc1"), (3, None, b"123"), (1, 2, b"bc")],
)
def test_stub_download_as_string(start, end, expected):
    """
    In this case, need to test the tests to make sure the implementation of
    download_as_string functions the same as Google cloud, reads are 0-indexed and
    inclusive of endpoints.
    """
    result = stub_download_as_string(start, end)
    assert result == expected


def test_gs_file_read(readable_gs_file):
    assert readable_gs_file.read(3) == b"abc"
    assert readable_gs_file.read(3) == b"123"
    assert readable_gs_file.read(3) == b""


def test_gs_file_read_none(readable_gs_file):
    assert readable_gs_file.read() == b"abc123"
    assert readable_gs_file.read() == b""


def test_gs_file_read_amount_then_none(readable_gs_file):
    assert readable_gs_file.read(3) == b"abc"
    assert readable_gs_file.read() == b"123"
    assert readable_gs_file.read() == b""


def test_s3_file_key():
    s3_file = S3File(key="my_task", name="s3://foo/bar/baz.qux")
    assert s3_file.key == "bar/baz.qux"


def test_s3_file_bucket():
    s3_file = S3File(key="my_task", name="s3://foo/bar/baz.qux")
    assert s3_file.bucket == "foo"


def test_s3_file_md5sum_etag_is_md5sum(mocker):
    mocker.patch(
        "accession.file.S3File._get_md5sum_from_object_tagging", return_value=None
    )
    mocker.patch(
        "accession.file.S3File.object_metadata",
        mocker.PropertyMock(
            return_value={"ETag": '"6640ff9ee51263e73c16cb84109365b3"'}
        ),
    )
    s3_file = S3File(key="my_task", name="s3://foo/bar/baz.qux")
    assert s3_file.md5sum == "6640ff9ee51263e73c16cb84109365b3"


def test_s3_file_md5sum_etag_is_not_md5sum(mocker):
    mocker.patch(
        "accession.file.S3File._get_md5sum_from_object_tagging", return_value=None
    )
    mocker.patch(
        "accession.file.S3File.object_metadata",
        mocker.PropertyMock(return_value={"ETag": "md5um-2"}),
    )
    mocker.patch("accession.file.S3File._calculate_md5sum", return_value="abc")
    s3_file = S3File(key="my_task", name="s3://foo/bar/baz.qux")
    assert s3_file.md5sum == "abc"


def test_s3_file_md5sum_from_object_tagging(mocker):
    mocker.patch(
        "accession.file.S3File._get_md5sum_from_object_tagging", return_value="foo"
    )
    s3_file = S3File(key="my_task", name="s3://foo/bar/baz.qux")
    assert s3_file.md5sum == "foo"


def test_s3_file_md5sum_from_portal(mocker, auth):
    mocker.patch("accession.file.S3File._get_md5sum_from_portal", return_value="12")
    s3_file = S3File(key="my_task", name="s3://encode-files/bar/baz.qux")
    assert s3_file._get_md5sum_from_portal() == "12"


def test_s3_file_get_md5sum_from_portal(mocker, auth):
    stub_response = mocker.Mock()
    stub_response.json.return_value = {"@graph": [{"md5sum": "123abc"}]}
    mocker.patch("requests.get", return_value=stub_response)
    s3_file = S3File(key="my_task", name="s3://encode-files/bar/baz.qux")
    assert s3_file._get_md5sum_from_portal() == "123abc"


def test_s3_file_get_md5sum_from_portal_no_results_raises(mocker, auth):
    stub_response = mocker.Mock()
    stub_response.json.return_value = {"@graph": []}
    mocker.patch("requests.get", return_value=stub_response)
    s3_file = S3File(key="my_task", name="s3://encode-files/bar/baz.qux")
    with pytest.raises(ValueError):
        s3_file._get_md5sum_from_portal()


def test_s3_file_calculate_md5sum(mocker):
    mock_body = mocker.Mock()
    mock_body.iter_chunks.return_value = [b"foo", b"bar"]
    mock_object = {"Body": mock_body}
    mocker.patch("accession.file.S3File.get_object", return_value=mock_object)
    s3_file = S3File(key="my_task", name="s3://foo/bar/baz.qux")
    assert s3_file._calculate_md5sum() == "3858f62230ac3c915f300c664312c63f"


def test_s3_file_get_md5sum_from_object_tagging(mocker):
    stub_client = mocker.Mock()
    stub_client.get_object_tagging.return_value = {
        "TagSet": [{"Key": "md5sum", "Value": "123"}]
    }
    mocker.patch(
        "accession.file.S3File.client", mocker.PropertyMock(return_value=stub_client)
    )
    s3_file = S3File(key="my_task", name="s3://foo/bar/baz.qux")
    assert s3_file._get_md5sum_from_object_tagging() == "123"


def test_s3_file_get_md5sum_from_object_tagging_md5sum_not_in_tagging(mocker):
    stub_client = mocker.Mock()
    stub_client.get_object_tagging.return_value = {"TagSet": []}
    mocker.patch(
        "accession.file.S3File.client", mocker.PropertyMock(return_value=stub_client)
    )
    s3_file = S3File(key="my_task", name="s3://foo/bar/baz.qux")
    assert s3_file._get_md5sum_from_object_tagging() is None


def test_s3_file_size(mocker):
    mocker.patch(
        "accession.file.S3File.object_metadata",
        mocker.PropertyMock(return_value={"ContentLength": 123}),
    )
    s3_file = S3File(key="my_task", name="s3://foo/bar/baz.qux")
    assert s3_file.size == 123


def test_s3_file_get_uri_without_scheme():
    s3_file = S3File(key="my_task", name="s3://foo/bar/baz.qux")
    assert s3_file.get_uri_without_scheme() == "foo/bar/baz.qux"


def test_local_file_md5sum(mocker, local_file):
    mocker.patch("builtins.open", mocker.mock_open(read_data=b"foo"))
    assert local_file.md5sum == "acbd18db4cc2f85cedef654fccc4a4d8"


def test_local_file_size(mocker, local_file):
    stat_result = mocker.Mock(st_size=3)
    mocker.patch("pathlib.Path.stat", return_value=stat_result)
    assert local_file.size == 3


def test_local_file_get_uri_without_scheme(local_file):
    assert local_file.get_uri_without_scheme() == "bar"


@pytest.mark.filesystem
def test_local_file_read_whole_file(test_file, test_file_contents):
    assert test_file.read() == test_file_contents
    assert test_file.read() == b""


@pytest.mark.filesystem
def test_local_file_read_chunks(test_file):
    assert test_file.read(num_bytes=6) == b"foo ba"
    assert test_file.read(num_bytes=5) == b"r baz"
    assert test_file.read(num_bytes=5) == b""


@pytest.mark.filesystem
def test_local_file_read_chunks_then_all(test_file):
    assert test_file.read(num_bytes=6) == b"foo ba"
    assert test_file.read() == b"r baz"
    assert test_file.read() == b""


@pytest.mark.filesystem
def test_local_file_read_bytes(test_file, test_file_contents):
    assert test_file.read_bytes() == test_file_contents


@pytest.mark.filesystem
def test_local_file_read_json(tmp_path):
    path = tmp_path / "test_file.json"
    path.write_bytes(b'{"foo":"bar"}')
    file = LocalFile(key="foo", name=str(path))
    result = file.read_json()
    assert result == {"foo": "bar"}



import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--docker",
        action="store",
        default="",
        help=(
            "Docker image to use for local encoded server in tests, e.g. quay.io/encode"
            "-dcc/accession:v80.0"
        ),
    )


@pytest.fixture(scope="session")
def encoded_docker_image(request):
    return request.config.getoption("--docker")



from accession.database.models import QualityMetric, RunStatus, WorkflowLabel


def test_run_status_str():
    assert str(RunStatus.Succeeded) == "succeeded"


def test_workflow_label_str():
    assert repr(WorkflowLabel(key="foo", value="bar")) == "bar"


def test_quality_metric_repr():
    assert repr(QualityMetric(portal_at_id="foo")) == "foo"



import calendar
import csv
import datetime
import re
import sys
from abc import ABC, abstractclassmethod, abstractmethod
from enum import Enum, auto
from itertools import zip_longest
from typing import List, Optional, Sequence, Tuple, Type, TypeVar, Union

from sqlalchemy.orm import Query

from accession.database.models import Run
from accession.database.query import DbQuery
from accession.helpers import flatten

# Kind of ugly, see https://stackoverflow.com/a/44644576/13501789
_WorkflowId = TypeVar("_WorkflowId", bound="WorkflowId")
_CromwellWorkflowLabel = TypeVar(
    "_CromwellWorkflowLabel", bound="CromwellWorkflowLabel"
)
_ExperimentId = TypeVar("_ExperimentId", bound="ExperimentId")

# Validate dates in month/day/year form, where month & day are either one or two digits
# and year is either two or four digits
DATE_REGEX = r"^(?P<month>\d{1,2})(\/(?P<day>\d{1,2}))?(\/(?P<year>\d{2}|\d{4}))?$"

# Assume we will only accesion things in the 21st century
CURRENT_CENTURY_PREFIX = "20"


class AbstractId(ABC):
    @abstractclassmethod
    def from_string(cls, id: str) -> "Optional[AbstractId]":
        raise NotImplementedError

    @abstractmethod
    def get_query(self, db_query: DbQuery) -> Query:
        raise NotImplementedError


class WorkflowId(AbstractId):
    REGEX = r"^[a-f\d]{8}-([a-f\d]{4}-){3}[a-f\d]{12}$"

    def __init__(self, workflow_id: str) -> None:
        self.workflow_id = workflow_id

    @classmethod
    def from_string(cls: Type[_WorkflowId], workflow_id: str) -> Optional[_WorkflowId]:
        """
        Tries to parse the provided string into a `WorkflowId` and returns it, otherwise
        if it cannot be parsd returns `None`.
        """
        result = re.match(cls.REGEX, workflow_id)
        if result is not None:
            return cls(workflow_id)
        return None

    def get_query(self, db_query: DbQuery) -> Query:
        return db_query.get_runs_by_workflow_id(self.workflow_id)


class CromwellWorkflowLabel(AbstractId):
    """
    Represents a Cromwell workflow label (key value pair). "Cromwell" is in the name to
    avoid potential aliasing with `accession.database.models.WorkflowLabel`
    """

    CAPER_STR_LABEL = "caper-str-label"

    def __init__(self, key: str, value: str) -> None:
        self.key = key
        self.value = value

    @classmethod
    def from_string(
        cls: Type[_CromwellWorkflowLabel], workflow_label: str
    ) -> _CromwellWorkflowLabel:
        """
        If given a key=value pair extracts the key and value, otherwise uses
        `cls.CAPER_STR_LABEL` as the key and uses everything else as the value.

        `=` is valid in keys or values, users need to escape any `=` that in the key or
        value with a backslash so that they don't get split on. Here we check that there
        is only one non-escaped equals sign, and raise an error if that is not the case.

        Unlike the other classes `WorkflowId` and `ExperimentId`, this will not return
        `None`, since almost any string will be considered as valid here.
        """
        count_equals = workflow_label.count("=")
        count_escaped_equals = workflow_label.count("\\=")

        if count_equals - count_escaped_equals == 0:
            return cls(cls.CAPER_STR_LABEL, workflow_label)

        if count_equals - count_escaped_equals != 1:
            raise ValueError(
                "Found more than one unescaped `=` in key=value pair, must only '"
                "specify one so parsing is not ambiguous"
            )

        for i, char in enumerate(workflow_label):
            if char == "=":
                if workflow_label[i - 1] != "\\":
                    key, value = workflow_label[0:i], workflow_label[i + 1 :]
                    return cls(key, value)

        # Can skip coverage here, we know the loop above always executes on a string
        # with one non-escaped equals sign in it
        raise ValueError("Could not detect key-value pair")  # pragma: no cover

    def get_query(self, db_query: DbQuery) -> Query:
        return db_query.get_runs_by_workflow_label(key=self.key, value=self.value)


class ExperimentId(AbstractId):
    REGEX = r"^(\/?experiments\/)?(?P<accession>ENCSR\d{3}[A-Z]{3})\/?$"

    def __init__(self, experiment_accession: str) -> None:
        self.experiment_id = f"/experiments/{experiment_accession}/"

    @classmethod
    def from_string(
        cls: Type[_ExperimentId], experiment_id: str
    ) -> Optional[_ExperimentId]:
        """
        Tries to parse the provided string into a `WorkflowId` and returns it, otherwise
        if it cannot be parsed returns `None`.
        """
        result = re.match(cls.REGEX, experiment_id)
        if result is not None:
            return cls(result.group("accession"))
        return None

    def get_query(self, db_query: DbQuery) -> Query:
        return db_query.get_runs_by_experiment_at_id(self.experiment_id)


class FillStrategy(Enum):
    """
    Helper enum that controls how ambiguous day ranges in dates are resolved
    """

    Min = auto()
    Max = auto()


def info(
    db_query: DbQuery, ids: Optional[List[str]] = None, date_range: Optional[str] = None
) -> None:
    """
    The main driver for the subcommand, organizing input parsing, querying the DB, and
    printing the results.
    """
    results = make_queries(db_query, ids, date_range)
    rows = get_rows_from_runs(results)
    rows = add_header_to_rows(rows)
    print_rows(rows)


def make_queries(
    db_query: DbQuery, ids: Optional[List[str]] = None, date_range: Optional[str] = None
) -> List[Run]:
    """
    Transforms user parameters into queries, executes them with additional filtering and
    aggregation, and returns the query results.
    """
    if ids is not None and ids:
        parsed_ids = parse_ids(ids)
        queries = [i.get_query(db_query) for i in parsed_ids]
        if len(queries) > 1:
            query_union = queries[0].union(*queries[1:])
        else:
            query_union = queries[0]
    else:
        query_union = db_query.get_all_runs()

    if date_range is not None:
        start, end = parse_date_range(date_range)
        filtered = DbQuery.filter_results_by_date_range(query_union, start, end)
        return filtered.all()
    else:
        return query_union.all()


def parse_ids(
    ids: List[str]
) -> List[Union[WorkflowId, CromwellWorkflowLabel, ExperimentId]]:
    """
    Simple wrapper to parse each id in the list into the appropriate type
    """
    return [parse_id(i) for i in ids]


def parse_id(id: str) -> Union[WorkflowId, CromwellWorkflowLabel, ExperimentId]:
    """
    First we try to parse as a workflow ID, since this is the most specific and has a
    known regex pattern. Then we try to parse as an experiment accession or `@id`, which
    also has a known pattern but is less general. Failing both, we interpret as a string
    label, since this can be a nearly arbitrary string, which only succeeds or raises.
    """
    workflow_id = WorkflowId.from_string(id)
    if workflow_id is not None:
        return workflow_id
    experiment_id = ExperimentId.from_string(id)
    if experiment_id is not None:
        return experiment_id
    workflow_label = CromwellWorkflowLabel.from_string(id)
    return workflow_label


def parse_date_range(
    date_range: str
) -> Tuple[Optional[datetime.date], Optional[datetime.date]]:
    """
    Returns a (start, end) tuple. If a bare date is specified, i.e. no dashes, then
    returns the same date for start and end if the day of the month is provided, or if
    the day of the month is not provided then will return a range spanning the entire
    month. If a leading `\\-` is specified, then returns `None` as the start. If a
    trailing `-` is specified, then `None` is returned as the end.

    Specifying two or more dashes anywhere in the pattern will raise a ValueError.
    Specifying a range that resolves in such a manner that start > end will also result
    in a ValueError.
    """
    date_range = date_range.lstrip("\\")
    split = date_range.split("-")
    if len(split) > 2:
        raise ValueError(
            "Invalid date range, should only specify one dash in date range string."
        )
    if len(split) == 1:
        date = split[0]
        return parse_date(date), parse_date(date, fill_strategy=FillStrategy.Max)
    if not split[0]:
        start = None
    else:
        start = parse_date(split[0])
    if not split[1]:
        end = None
    else:
        end = parse_date(split[1], fill_strategy=FillStrategy.Max)
    if start is not None and end is not None:
        if start > end:
            raise ValueError("Start date must be less than or equal to end date.")
    return (start, end)


def parse_date(
    date: str, fill_strategy: FillStrategy = FillStrategy.Min
) -> datetime.date:
    """
    Parses date string according to regex. The regex is such that we always recieve a
    month if the string validates. If year is not specified then we use the current
    year. If we don't recieve a day then we choose either the first day of the month
    if the fill_strategy is FillStrategy.Min, or the last day of the month if using
    FillStrategy.Max.
    """
    parsed = re.match(DATE_REGEX, date)
    if parsed is None:
        raise ValueError(f"Could not parse date {date}")

    month = int(parsed.group("month"))

    parsed_year = parsed.group("year")
    if parsed_year is None:
        year = get_today().year
    elif len(parsed_year) == 2:
        year = int(CURRENT_CENTURY_PREFIX + parsed_year)
    else:
        year = int(parsed_year)

    parsed_day = parsed.group("day")
    if parsed_day is None:
        if fill_strategy == FillStrategy.Min:
            # The first day of the month is always 1
            day = 1
        elif fill_strategy == FillStrategy.Max:
            _, day = calendar.monthrange(year, month)
    else:
        day = int(parsed_day)

    return datetime.date(month=month, day=day, year=year)


def get_today() -> datetime.date:
    """
    Keep as separate user defined function to make mocking parse_date easier. See
    https://nedbatchelder.com/blog/201209/mocking_datetimetoday.html
    """
    return datetime.date.today()


def expand_row(
    row: Sequence[Union[str, Sequence[Union[str, Sequence[str]]]]]
) -> List[List[str]]:
    """
    Converts a row with lists as elements into multiple rows with empty strings
    inserted to make column widths more consistent when eventually printing. This is
    easiest to explain visually. A list like `["foo", ["bar", "baz"]]` would be turned
    into `[["foo", "bar"], ["", "baz"]]`. This is used to align the files in the run
    tables.

    See https://mypy.readthedocs.io/en/latest/common_issues.html#invariance-vs-covariance
    for details on why we need Sequence in the signature.
    """
    elems_as_lists = []
    for elem in row:
        if isinstance(elem, list):
            elems_as_lists.append(elem)
        else:
            elems_as_lists.append([elem])
    aligned = [list(i) for i in zip_longest(*elems_as_lists, fillvalue="")]
    return aligned


def get_rows_from_runs(runs: List[Run]) -> List[List[str]]:
    """
    We don't recursively flatten the quality metrics list for each file, otherwise the
    table would take up a lot of vertical space than necessary.
    """
    rows: List[List[str]] = []
    for run in runs:
        run_files = [
            [file.portal_at_id, str(file.status), str(file.quality_metrics).strip("[]")]
            for file in run.files
        ]
        formatted = expand_row(
            [
                run.experiment_at_id,
                run.workflow_id,
                str(run.workflow_labels).strip("[]"),
                str(run.status),
                run.date_created.strftime("%m/%d/%y"),
                run_files,
            ]
        )
        rows.extend(formatted)
    return [flatten(row) for row in rows]


def add_header_to_rows(rows: List[List[str]]) -> List[List[str]]:
    header = [
        [
            "Experiment @id",
            "WF ID",
            "WF Labels",
            "Accessioning Status",
            "Date",
            "File @id",
            "File status",
            "Quality Metrics",
        ]
    ]
    return header + rows


def print_rows(rows: List[List[str]]) -> None:
    """
    Write rows in tabular format to stdout. Unfortunately the `lineterminator` kwarg
    seems to be necessary, otherwise the writer will use `\r\n` for newlines. Since we
    are writing to sys.stdout we can't configure it with the recommended `newline=""` as
    per the `csv` docs.
    """
    writer = csv.writer(sys.stdout, delimiter="\t", lineterminator="\n")
    for row in rows:
        writer.writerow(row)



import datetime
from typing import Optional

from sqlalchemy import and_
from sqlalchemy.orm import Query, Session

from accession.database.connection import DbSession
from accession.database.models import Run, WorkflowLabel


class DbQuery:
    def __init__(self, session: DbSession):
        self._session = session

    @property
    def session(self) -> Session:
        """
        Avoids an extra indirection by forwarding the session so we don't have to do
        `self.session.session.foo()` every time
        """
        return self._session.session

    def get_runs_by_experiment_at_id(self, experiment_at_id: str) -> Query:
        results = self.session.query(Run).filter_by(experiment_at_id=experiment_at_id)
        return results

    def get_runs_by_workflow_id(self, workflow_id: str) -> Query:
        results = self.session.query(Run).filter_by(workflow_id=workflow_id)
        return results

    def get_runs_by_workflow_label(self, key: str, value: str) -> Query:
        """
        Workflow labels are key-value pairs. The most common form this takes is
        ("caper-str-label", "VALUE"). Rather than overload this function we assume we
        have already determined if the label is a `caper-str-label` label or not
        """
        results = (
            self.session.query(Run)
            .join(WorkflowLabel)
            .filter(and_(WorkflowLabel.key == key, WorkflowLabel.value == value))
        )
        return results

    def get_runs_by_date_range(
        self, start: Optional[datetime.date] = None, end: Optional[datetime.date] = None
    ) -> Query:
        """
        Search for runs between the specified start and end time, inclusive of
        endpoints. If `start` is not specified, will search from the beginning of time,
        and if `end` is not specified then will search to the end of time.
        """
        results = self.filter_results_by_date_range(self.session.query(Run), start, end)
        return results

    def get_all_runs(self) -> Query:
        return self.session.query(Run)

    @staticmethod
    def filter_results_by_date_range(
        results: Query,
        start: Optional[datetime.date] = None,
        end: Optional[datetime.date] = None,
    ) -> Query:
        if start is None:
            start = datetime.date.min
        if end is None:
            end = datetime.date.max
        if start > end:
            raise ValueError("Start time must be less than end")
        filtered = results.filter(Run.date_created >= start).filter(
            Run.date_created <= end
        )
        return filtered



import hashlib
import io
import json
from abc import ABC, abstractmethod
from base64 import b64decode
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional
from urllib.parse import urlparse

import boto3
import requests
from google.cloud.storage.blob import Blob
from google.cloud.storage.client import Client

from accession.helpers import get_api_keys_from_env
from accession.task import Task

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
    from mypy_boto3_s3.type_defs import GetObjectOutputTypeDef, HeadObjectOutputTypeDef


class File(ABC):
    """
    Abstract base class defining the interface for different concrete file classes to
    implement.
    """

    @abstractmethod
    def __init__(
        self,
        key: str,
        name: str,
        task: Optional[Task] = None,
        used_by_task: Optional[Task] = None,
    ) -> None:
        self.filename = name
        self.filekeys = [key]
        self.task = task
        self.used_by_tasks = [used_by_task] if used_by_task else []

    @property
    @abstractmethod
    def md5sum(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def size(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_uri_without_scheme(self) -> str:
        """
        String representing the scheme for URIs for files represented by the class.
        """
        raise NotImplementedError("Derived classes should provide their own SCHEMEs")

    @abstractmethod
    def read_bytes(self) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def read_json(self) -> Dict[str, Any]:
        raise NotImplementedError

    def get_task(self) -> Task:
        """
        Either gets the `File`'s task or raises a ValueError.
        """
        task = self.task
        if task is None:
            raise ValueError("No task found")
        return task

    def get_filename_for_encode_alias(self) -> str:
        return self.get_uri_without_scheme().replace("/", "-")


class GSFile(File):
    """
    Wrapper around GCS blob class to better map to portal metadata and provide a read()
    interface for in-memory transfer to s3
    """

    SCHEME = "gs://"

    def __init__(
        self,
        key: str,
        name: str,
        task: Optional[Task] = None,
        used_by_task: Optional[Task] = None,
        client: Optional[Client] = None,
    ) -> None:
        """
        Initializes self.pos to 0 for keeping track of number of bytes read from file.
        """
        super().__init__(key, name, task, used_by_task)
        self.pos = 0
        self._blob: Optional[Blob] = None
        self._client: Optional[Client] = client

    @property
    def blob(self) -> Blob:
        if self._blob is None:
            blob = Blob.from_string(self.filename, client=self.client)
            blob.reload()
            self._blob = blob
        return self._blob

    @property
    def client(self) -> Client:
        if self._client is None:
            client = Client()
            self._client = client
        return self._client

    @property
    def md5sum(self) -> str:
        """
        Returns md5sum of the file in hex. Need to wrap around gcloud API's md5sums,
        which are returned as base64, to match ENCODE portal md5sums.
        """
        return self.b64_to_hex(self.blob.md5_hash)

    @property
    def size(self) -> int:
        return self.blob.size

    @staticmethod
    def b64_to_hex(value: str) -> str:
        return b64decode(value).hex()

    def get_uri_without_scheme(self) -> str:
        return f"{self.blob.bucket.name}/{self.blob.name}"

    def read(self, num_bytes: Optional[int] = None) -> bytes:
        """
        `Blob.download_as_string()` takes `start` and `end` kwargs to specify a byte
        range. These are 0-indexed and inclusive of endpoints. If the position is
        greater than or equal to the size of the object then we treat that as EOF and
        return an empty byte string `b''`. As per Python convention, when read() is
        called with no read size then the remainder of the file is returned.

        See https://googleapis.dev/python/storage/latest/blobs.html#google.cloud.Blob.download_as_string
        """
        if self.pos >= self.size:
            read_bytes = b""
        else:
            if num_bytes is None:
                read_bytes = self.blob.download_as_string(start=self.pos)
                self.pos += len(read_bytes)
            else:
                read_bytes = self.blob.download_as_string(
                    start=self.pos, end=self.pos + num_bytes - 1
                )
                self.pos += num_bytes
        return read_bytes

    def read_bytes(self) -> bytes:
        """
        Downloads file as bytes. We bypass the `read` interface so we don't mess with
        `self.pos`.
        """
        return self.blob.download_as_string()

    def read_json(self) -> Dict[str, Any]:
        """
        Read file and convert to JSON
        """
        return json.loads(self.read_bytes().decode())


class S3File(File):
    """
    For s3 to s3 copy we can just pass the StreamingBody object returned by `get_object`
    so the `read` method is not needed on this class.
    """

    SCHEME = "s3://"
    MD5_CHUNKSIZE = 8192
    MD5SUM_TAG_KEY = "md5sum"
    PORTAL_BUCKETS = (
        "encode-public",
        "encode-private",
        "encode-files",
        "encode-files-dev",
    )

    def __init__(
        self,
        key: str,
        name: str,
        task: Optional[Task] = None,
        used_by_task: Optional[Task] = None,
        client: Optional["S3Client"] = None,
    ) -> None:
        """
        Initializes self.pos to 0 for keeping track of number of bytes read from file.
        """
        super().__init__(key, name, task, used_by_task)
        self._md5sum: Optional[str] = None
        self._object_metadata: Optional["HeadObjectOutputTypeDef"] = None
        self._client: Optional["S3Client"] = client

    @property
    def object_metadata(self) -> "HeadObjectOutputTypeDef":
        """
        Cache object metadata to avoid repeated calls to GetObject
        """
        if self._object_metadata is None:
            self._object_metadata = self.client.head_object(
                Bucket=self.bucket, Key=self.key
            )
        return self._object_metadata

    @property
    def client(self) -> "S3Client":
        if self._client is None:
            client = boto3.client("s3")
            self._client = client
        return self._client

    @property
    def bucket(self) -> str:
        return urlparse(self.filename, allow_fragments=False).netloc

    @property
    def key(self) -> str:
        return urlparse(self.filename, allow_fragments=False).path.lstrip("/")

    @property
    def md5sum(self) -> str:
        """
        On S3 the ETag is the md5sum if the file was not uploaded with multipart upload.
        If multipart upload was not used then we can tell from the `-` in the ETag and
        we need to calculate the md5sums ourselves.

        MD5 calculation is also bypassed if there is an "md5sum" key in the object tags
        """
        if self._md5sum is None:
            if self.bucket in self.PORTAL_BUCKETS:
                self._md5sum = self._get_md5sum_from_portal()
                return self._md5sum
            md5sum_from_tagging = self._get_md5sum_from_object_tagging()
            if md5sum_from_tagging is not None:
                self._md5sum = md5sum_from_tagging
                return md5sum_from_tagging
            # ETag is wrapped in quotes for some reason
            etag = self.object_metadata["ETag"].strip('"')
            if len(etag) == 32 and "-" not in etag:
                self._md5sum = etag
            else:
                self._md5sum = self._calculate_md5sum()
        return self._md5sum

    @property
    def size(self) -> int:
        return self.object_metadata["ContentLength"]

    def _calculate_md5sum(self) -> str:
        """
        If file was uploaded via multipart then Etag is not the md5sum
        """
        md5_hash = hashlib.md5()
        for chunk in self.get_object()["Body"].iter_chunks(self.MD5_CHUNKSIZE):
            md5_hash.update(chunk)
        return md5_hash.hexdigest()

    def _get_md5sum_from_object_tagging(self) -> Optional[str]:
        """
        Tries to pull md5sum from the object tags, as is the case if the md5sum was
        precalculated with https://github.com/ENCODE-DCC/s3-md5-hash.
        """
        tagging = self.client.get_object_tagging(Bucket=self.bucket, Key=self.key)
        md5sum_tag = [
            tag for tag in tagging["TagSet"] if tag["Key"] == self.MD5SUM_TAG_KEY
        ]
        if not md5sum_tag:
            return None
        return md5sum_tag[0]["Value"]

    def _get_md5sum_from_portal(self) -> str:
        """
        Will error if the md5sum could not be found on the portal, most likely to occur
        when file is not in ES for some reason.
        """
        auth = get_api_keys_from_env()
        response = requests.get(
            f"https://www.encodeproject.org/search/?type=File&field=md5sum&s3_uri={self.filename}",
            auth=auth,
        )
        response.raise_for_status()
        data = response.json()
        if len(data["@graph"]) != 1:
            raise ValueError(
                f"Could not find file on portal with s3_uri {self.filename}"
            )
        return data["@graph"][0]["md5sum"]

    def get_object(self) -> "GetObjectOutputTypeDef":
        return self.client.get_object(Bucket=self.bucket, Key=self.key)

    def get_uri_without_scheme(self) -> str:
        return f"{self.bucket}/{self.key}"

    def read_bytes(self) -> bytes:
        """
        Download the whole file as bytes
        """
        buffer = io.BytesIO()
        self.client.download_fileobj(Bucket=self.bucket, Key=self.key, Fileobj=buffer)
        return buffer.getvalue()

    def read_json(self) -> Dict[str, Any]:
        return json.loads(self.read_bytes().decode())


class LocalFile(File):
    MD5_CHUNKSIZE = 8192

    def __init__(
        self,
        key: str,
        name: str,
        task: Optional[Task] = None,
        used_by_task: Optional[Task] = None,
    ) -> None:
        """
        Initializes self.pos to 0 for keeping track of number of bytes read from file.
        """
        super().__init__(key, name, task, used_by_task)
        self.pos = 0
        self._md5sum: Optional[str] = None
        self._size: Optional[int] = None

    @property
    def md5sum(self) -> str:
        if self._md5sum is None:
            md5_hash = hashlib.md5()
            with open(self.filename, "rb") as f:
                for chunk in iter(lambda: f.read(self.MD5_CHUNKSIZE), b""):
                    md5_hash.update(chunk)
            self._md5sum = md5_hash.hexdigest()
        return self._md5sum

    @property
    def size(self) -> int:
        if self._size is None:
            self._size = Path(self.filename).stat().st_size
        return self._size

    def get_uri_without_scheme(self) -> str:
        return self.filename

    def read(self, num_bytes: Optional[int] = None) -> bytes:
        if self.pos >= self.size:
            return b""

        with open(self.filename, "rb") as f:
            f.seek(self.pos)
            if num_bytes is None:
                read_bytes = f.read()
                self.pos += len(read_bytes)
            else:
                read_bytes = f.read(num_bytes)
                self.pos += num_bytes
        return read_bytes

    def read_bytes(self) -> bytes:
        """
        Downloads file as bytes. We bypass the `read` interface so we don't mess with
        `self.pos`.
        """
        with open(self.filename, "rb") as f:
            return f.read()

    def read_json(self) -> Dict[str, Any]:
        """
        Read file and convert to JSON
        """
        return json.loads(self.read_bytes().decode())



import datetime
from contextlib import suppress as does_not_raise

import pytest

from accession.database.models import (
    DbFile,
    QualityMetric,
    Run,
    RunStatus,
    WorkflowLabel,
)
from accession.encode_models import FileStatus
from accession.subcommands.info import (
    CromwellWorkflowLabel,
    ExperimentId,
    FillStrategy,
    WorkflowId,
    expand_row,
    get_rows_from_runs,
    make_queries,
    parse_date,
    parse_date_range,
    parse_id,
    parse_ids,
    print_rows,
)


def test_workflow_id_from_string_valid_workflow_id():
    workflow_id = "60fb02a7-31ae-4e9e-8e77-9c2722189cb4"
    result = WorkflowId.from_string(workflow_id)
    assert result.workflow_id == workflow_id


def test_workflow_id_from_string_invalid_workflow_id_returns_none():
    workflow_id = "foo"
    result = WorkflowId.from_string(workflow_id)
    assert result is None


def test_workflow_label_from_string():
    label = "key=value"
    result = CromwellWorkflowLabel.from_string(label)
    assert result.key == "key"
    assert result.value == "value"


def test_workflow_label_from_string_with_escaped_equals_in_key():
    label = "k\\=ey=value"
    result = CromwellWorkflowLabel.from_string(label)
    assert result.key == "k\\=ey"
    assert result.value == "value"


def test_workflow_label_from_string_no_key():
    label = "val\\=ue"
    result = CromwellWorkflowLabel.from_string(label)
    assert result.key == "caper-str-label"
    assert result.value == "val\\=ue"


def test_workflow_label_from_string_multiple_equals_raises():
    label = "key=val=valu\\=e"
    with pytest.raises(ValueError):
        CromwellWorkflowLabel.from_string(label)


@pytest.mark.parametrize(
    "experiment_id",
    ["ENCSR123ABC", "/experiments/ENCSR123ABC/", "experiments/ENCSR123ABC"],
)
def test_experiment_id_from_string(experiment_id):
    result = ExperimentId.from_string(experiment_id)
    assert result.experiment_id == "/experiments/ENCSR123ABC/"


def test_experiment_id_from_string_invalid_returns_none():
    result = ExperimentId.from_string("foo")
    assert result is None


def test_make_queries_labels_and_date_range(db_query, run, additional_runs):
    results = make_queries(
        db_query,
        ids=["foo=bar", "60fb02a7-31ae-4e9e-8e77-9c2722189cb4", "ENCSR123ABC"],
        date_range="3/10/20-3/11/20",
    )
    assert len(results) == 2


def test_make_queries_labels_no_date_range(db_query, run, additional_runs):
    results = make_queries(
        db_query, ids=["foo=bar", "60fb02a7-31ae-4e9e-8e77-9c2722189cb4", "ENCSR123ABC"]
    )
    assert len(results) == 3


def test_make_queries_no_labels_date_range(db_query, run, additional_runs):
    results = make_queries(db_query, date_range="3/11/20")
    assert len(results) == 1


def test_parse_ids():
    ids = ["ENCSR123ABC", "key=value", "60fb02a7-31ae-4e9e-8e77-9c2722189cb4"]
    result = parse_ids(ids)
    assert isinstance(result[0], ExperimentId)
    assert isinstance(result[1], CromwellWorkflowLabel)
    assert isinstance(result[2], WorkflowId)


def test_parse_id_workflow_id():
    workflow_id = "60fb02a7-31ae-4e9e-8e77-9c2722189cb4"
    result = parse_id(workflow_id)
    assert isinstance(result, WorkflowId)


def test_parse_id_experiment_id():
    experiment_id = "ENCSR123ABC"
    result = parse_id(experiment_id)
    assert isinstance(result, ExperimentId)


def test_parse_id_workflow_label():
    workflow_label = "label"
    result = parse_id(workflow_label)
    assert isinstance(result, CromwellWorkflowLabel)


def test_parse_date_range_no_start(mocker):
    m = mocker.MagicMock()
    mocker.patch("accession.subcommands.info.parse_date", m)
    start, _ = parse_date_range("\\-3/4/57")
    assert start is None
    m.assert_called_once_with("3/4/57", fill_strategy=FillStrategy.Max)


def test_parse_date_range_no_end(mocker):
    m = mocker.MagicMock()
    mocker.patch("accession.subcommands.info.parse_date", m)
    _, end = parse_date_range("3/4/57-")
    assert end is None
    m.assert_called_once_with("3/4/57")


def test_parse_date_range(mocker):
    m = mocker.MagicMock(side_effect=(3, 4))
    mocker.patch("accession.subcommands.info.parse_date", m)
    parse_date_range("3/4/57-3/5/57")
    m.assert_has_calls(
        [mocker.call("3/4/57"), mocker.call("3/5/57", fill_strategy=FillStrategy.Max)]
    )


def test_parse_date_range_start_greater_than_end_raises(mocker):
    """
    The side effect here determines the fake values of start and end that we "parse" out
    """
    m = mocker.MagicMock(side_effect=(4, 3))
    mocker.patch("accession.subcommands.info.parse_date", m)
    with pytest.raises(ValueError):
        parse_date_range("3/5/57-3/4/57")


@pytest.mark.parametrize(
    "condition,date,expected",
    [
        (does_not_raise(), "4", datetime.date(2020, month=4, day=1)),
        (does_not_raise(), "4/5", datetime.date(2020, month=4, day=5)),
        (does_not_raise(), "04/05", datetime.date(2020, month=4, day=5)),
        (does_not_raise(), "4/5/19", datetime.date(2019, month=4, day=5)),
        (does_not_raise(), "4/5/1919", datetime.date(1919, month=4, day=5)),
        (pytest.raises(ValueError), "4/5/", None),
        (pytest.raises(ValueError), "today", None),
    ],
)
def test_parse_date(mocker, condition, date, expected):
    mocker.patch(
        "accession.subcommands.info.get_today", return_value=datetime.date(2020, 6, 7)
    )
    with condition:
        result = parse_date(date)
        assert result == expected


def test_parse_date_fillstrategy_max(mocker):
    """
    April always has 30 days, so this is OK.
    """
    mocker.patch(
        "accession.subcommands.info.get_today", return_value=datetime.date(2020, 6, 7)
    )
    result = parse_date("4", fill_strategy=FillStrategy.Max)
    assert result == datetime.date(2020, month=4, day=30)


@pytest.mark.parametrize(
    "input,expected",
    [
        ([1, [2, 3]], [[1, 2], ["", 3]]),
        ([1, [[2, 3], [4, 5]]], [[1, [2, 3]], ["", [4, 5]]]),
    ],
)
def test_expand_row(input, expected):
    result = expand_row(input)
    assert result == expected


def test_get_rows_from_runs():
    run = Run(
        experiment_at_id="foo",
        workflow_id="123",
        date_created=datetime.date(1999, 9, 9),
        status=RunStatus.Succeeded,
        workflow_labels=[WorkflowLabel(key="foo", value="bar")],
        files=[
            DbFile(
                portal_at_id="baz",
                status=FileStatus.Released,
                quality_metrics=[QualityMetric(portal_at_id="qux")],
            ),
            DbFile(portal_at_id="quux", status=FileStatus.Released, quality_metrics=[]),
        ],
    )
    runs = [run]
    result = get_rows_from_runs(runs)
    assert result == [
        ["foo", "123", "bar", "succeeded", "09/09/99", "baz", "released", "qux"],
        ["", "", "", "", "", "quux", "released", ""],
    ]


def test_print_rows(capsys):
    print_rows([["foo", "bar"], ["baz", "qux"]])
    captured = capsys.readouterr()
    assert captured.out == "foo\tbar\nbaz\tqux\n"



import datetime
import enum  # We don't `import enum from Enum` to avoid name conflict with sqlalchemy.Enum
from typing import Type, TypeVar

from sqlalchemy import Column, Date, Enum, ForeignKey, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from accession.database import (
    FILE_TABLE_NAME,
    QUALITY_METRIC_TABLE_NAME,
    RUN_TABLE_NAME,
    WORKFLOW_LABEL_TABLE_NAME,
)
from accession.encode_models import EncodeFile, EncodeQualityMetric, FileStatus

_DbFile = TypeVar("_DbFile", bound="DbFile")
_QualityMetric = TypeVar("_QualityMetric", bound="QualityMetric")

Base = declarative_base()


class RunStatus(enum.Enum):
    Succeeded = "succeeded"
    Failed = "failed"

    def __str__(self) -> str:
        return self.value


class Run(Base):
    """
    Representation of an accessioning run of a single experiment. `workflow_labels` is
    somewhat special, ideally we would store them inline in an `ARRAY` but that's not
    supported in SQLite, so we need to use a foreign key relationship.
    """

    __tablename__ = RUN_TABLE_NAME
    id = Column(Integer, primary_key=True)
    experiment_at_id = Column(String, nullable=False)
    workflow_id = Column(String, nullable=False)
    date_created = Column(Date, nullable=False, default=datetime.date.today)
    status = Column(Enum(RunStatus), nullable=False)
    files = relationship("DbFile", backref=RUN_TABLE_NAME, uselist=True)
    workflow_labels = relationship(
        "WorkflowLabel", backref=RUN_TABLE_NAME, uselist=True
    )


class WorkflowLabel(Base):
    __tablename__ = WORKFLOW_LABEL_TABLE_NAME
    id = Column(Integer, primary_key=True)
    key = Column(String, nullable=False)
    value = Column(String, nullable=False)
    run = Column(Integer, ForeignKey(f"{RUN_TABLE_NAME}.id"))

    def __repr__(self) -> str:
        return self.value


class DbFile(Base):
    __tablename__ = FILE_TABLE_NAME
    id = Column(Integer, primary_key=True)
    portal_at_id = Column(String, nullable=False)
    status = Column(Enum(FileStatus), nullable=False)
    run = Column(Integer, ForeignKey(f"{RUN_TABLE_NAME}.id"))
    quality_metrics = relationship(
        "QualityMetric", backref=FILE_TABLE_NAME, uselist=True
    )

    @classmethod
    def from_encode_file(cls: Type[_DbFile], encode_file: EncodeFile) -> _DbFile:
        return cls(portal_at_id=encode_file.at_id, status=encode_file.status)


class QualityMetric(Base):
    __tablename__ = QUALITY_METRIC_TABLE_NAME
    id = Column(Integer, primary_key=True)
    portal_at_id = Column(String, nullable=False)
    file = Column(Integer, ForeignKey(f"{FILE_TABLE_NAME}.id"))

    @classmethod
    def from_encode_quality_metric(
        cls: Type[_QualityMetric], quality_metric: EncodeQualityMetric
    ) -> _QualityMetric:
        return cls(portal_at_id=quality_metric.at_id)

    def __repr__(self) -> str:
        return self.portal_at_id
